@@ -1,4 +1,4 @@
-// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -21,6 +21,14 @@
 #ifndef ROCPRIM_BENCHMARK_UTILS_HPP_
 #define ROCPRIM_BENCHMARK_UTILS_HPP_
 
+#include <benchmark/benchmark.h>
+
+// rocPRIM
+#include <rocprim/block/block_scan.hpp>
+#include <rocprim/device/config_types.hpp>
+#include <rocprim/device/detail/device_config_helper.hpp> // partition_config_params
+#include <rocprim/types.hpp>
+
 #include <algorithm>
 #include <iostream>
 #include <iterator>
@@ -31,11 +39,9 @@
 #include <sstream>
 #include <string>
 #include <type_traits>
+#include <utility>
 #include <vector>
 
-#include "benchmark/benchmark.h"
-#include <rocprim/rocprim.hpp>
-
 #define HIP_CHECK(condition)                                                                \
     {                                                                                       \
         hipError_t error = condition;                                                       \
@@ -50,6 +56,49 @@
 #define TUNING_SHARED_MEMORY_MAX 65536u
 // Support half operators on host side
 
+inline const char* get_seed_message()
+{
+    return "seed for input generation, either an unsigned integer value for determinisic results "
+           "or 'random' for different inputs for each repetition";
+}
+
+/// \brief Provides a sequence of seeds.
+class managed_seed
+{
+public:
+    /// \param[in] seed_string Either "random" to get random seeds,
+    ///   or an unsigned integer to get (a sequence) of deterministic seeds.
+    managed_seed(const std::string& seed_string)
+    {
+        is_random = seed_string == "random";
+        if(!is_random)
+        {
+            const unsigned int seed = std::stoul(seed_string);
+            std::seed_seq      seq{seed};
+            seq.generate(seeds.begin(), seeds.end());
+        }
+    }
+
+    unsigned int get_0() const
+    {
+        return is_random ? std::random_device{}() : seeds[0];
+    }
+
+    unsigned int get_1() const
+    {
+        return is_random ? std::random_device{}() : seeds[1];
+    }
+
+    unsigned int get_2() const
+    {
+        return is_random ? std::random_device{}() : seeds[2];
+    }
+
+private:
+    std::array<unsigned int, 3> seeds;
+    bool                        is_random;
+};
+
 ROCPRIM_HOST inline
 rocprim::native_half half_to_native(const rocprim::half& x)
 {
@@ -119,7 +168,7 @@ struct is_valid_for_int_distribution :
 template<typename Iterator>
 using it_value_t = typename std::iterator_traits<Iterator>::value_type;
 
-using engine_type = std::default_random_engine;
+using engine_type = std::minstd_rand;
 
 // generate_random_data_n() generates only part of sequence and replicates it,
 // because benchmarks usually do not need "true" random sequence.
@@ -170,15 +219,15 @@ inline auto generate_random_data_n(OutputIterator it,
 }
 
 template<class T>
-inline std::vector<T> get_random_data01(size_t size, float p, size_t max_random_size = 1024 * 1024)
+inline std::vector<T>
+    get_random_data01(size_t size, float p, unsigned int seed, size_t max_random_size = 1024 * 1024)
 {
-    engine_type gen{std::random_device{}()};
+    engine_type                 gen(seed);
     std::bernoulli_distribution distribution(p);
-    std::vector<T> data(size);
-    std::generate(
-        data.begin(), data.begin() + std::min(size, max_random_size),
-        [&]() { return distribution(gen); }
-    );
+    std::vector<T>              data(size);
+    std::generate(data.begin(),
+                  data.begin() + std::min(size, max_random_size),
+                  [&]() { return distribution(gen); });
     for(size_t i = max_random_size; i < size; i += max_random_size)
     {
         std::copy_n(data.begin(), std::min(size - i, max_random_size), data.begin() + i);
@@ -186,15 +235,6 @@ inline std::vector<T> get_random_data01(size_t size, float p, size_t max_random_
     return data;
 }
 
-template<class T>
-inline T get_random_value(T min, T max)
-{
-    T           result;
-    engine_type gen{std::random_device{}()};
-    generate_random_data_n(&result, 1, min, max, gen);
-    return result;
-}
-
 template<class T, class U = T>
 struct custom_type
 {
@@ -231,13 +271,63 @@ struct custom_type
     {
         return x == rhs.x && y == rhs.y;
     }
+
+    ROCPRIM_HOST_DEVICE custom_type& operator+=(const custom_type& rhs)
+    {
+        this->x += rhs.x;
+        this->y += rhs.y;
+        return *this;
+    }
 };
 
 template<typename>
-struct is_custom_type : std::false_type {};
+struct is_custom_type : std::false_type
+{};
 
 template<class T, class U>
-struct is_custom_type<custom_type<T,U>> : std::true_type {};
+struct is_custom_type<custom_type<T, U>> : std::true_type
+{};
+
+template<typename T, typename U>
+struct is_comparable
+{
+private:
+    // A dummy template function that attempts to compare two objects of types T and U
+    template<typename V, typename W>
+    static auto test(V&& v, W&& w)
+        -> decltype(std::declval<V>() < std::declval<W>(), std::true_type{});
+
+    // Fallback if the above template function is not valid
+    template<typename, typename>
+    static std::false_type test(...);
+
+public:
+    // Final result
+    static constexpr bool value = decltype(test<T, U>(std::declval<T>(), std::declval<U>()))::value;
+};
+
+template<typename T, typename U, typename V>
+struct is_comparable<custom_type<U, V>, T>
+    : std::conditional_t<rocprim::is_arithmetic<T>::value
+                             || !std::is_same<T, custom_type<U, V>>::value,
+                         std::false_type,
+                         std::true_type>
+{};
+
+template<class CustomType>
+struct custom_type_decomposer
+{
+    static_assert(is_custom_type<CustomType>::value,
+                  "custom_type_decomposer can only be used with instantiations of custom_type");
+
+    using T = typename CustomType::first_type;
+    using U = typename CustomType::second_type;
+
+    __host__ __device__ ::rocprim::tuple<T&, U&> operator()(CustomType& key) const
+    {
+        return ::rocprim::tuple<T&, U&>{key.x, key.y};
+    }
+};
 
 template<class OutputIterator, class Generator>
 inline auto generate_random_data_n(OutputIterator             it,
@@ -289,15 +379,111 @@ inline auto generate_random_data_n(OutputIterator             it,
 }
 
 template<class T, class U, class V>
-inline std::vector<T>
-    get_random_data(size_t size, U min, V max, size_t max_random_size = 1024 * 1024)
+inline std::vector<T> get_random_data(
+    size_t size, U min, V max, unsigned int seed, size_t max_random_size = 1024 * 1024)
 {
     std::vector<T> data(size);
-    engine_type    gen{std::random_device{}()};
+    engine_type    gen(seed);
     generate_random_data_n(data.begin(), size, min, max, gen, max_random_size);
     return data;
 }
 
+template<typename T, typename U>
+auto limit_cast(U value) -> T
+{
+    static_assert(rocprim::is_arithmetic<T>::value && rocprim::is_arithmetic<U>::value
+                      && is_comparable<T, U>::value,
+                  "Cannot use limit_cast with chosen types of T and U");
+
+    using common_type = typename std::common_type<T, U>::type;
+    if(rocprim::is_unsigned<T>::value)
+    {
+        if(value < 0)
+        {
+            return std::numeric_limits<T>::min();
+        }
+        if(static_cast<common_type>(value)
+           > static_cast<common_type>(std::numeric_limits<T>::max()))
+        {
+            return std::numeric_limits<T>::max();
+        }
+    }
+    else if(rocprim::is_signed<T>::value && rocprim::is_unsigned<U>::value)
+    {
+        if(value > std::numeric_limits<T>::max())
+        {
+            return std::numeric_limits<T>::max();
+        }
+    }
+    else if(rocprim::is_floating_point<T>::value)
+    {
+        return static_cast<T>(value);
+    }
+    else // Both T and U are signed
+    {
+        if(value < static_cast<common_type>(std::numeric_limits<T>::min()))
+        {
+            return std::numeric_limits<T>::min();
+        }
+        else if(value > static_cast<common_type>(std::numeric_limits<T>::max()))
+        {
+            return std::numeric_limits<T>::max();
+        }
+    }
+    return static_cast<T>(value);
+}
+
+// This overload below is selected for non-standard float types, e.g. half, which cannot be compared with the limit types.
+template<class T, class U, class V>
+inline auto limit_random_range(U range_start, V range_end)
+    -> std::enable_if_t<!is_custom_type<T>::value
+                            && (!is_comparable<T, U>::value || !is_comparable<T, V>::value),
+                        std::pair<T, T>>
+{
+    return {static_cast<T>(range_start), static_cast<T>(range_end)};
+}
+
+template<typename T, typename U, typename V>
+auto limit_random_range(U range_start, V range_end)
+    -> std::enable_if_t<(is_custom_type<T>::value && is_comparable<typename T::first_type, U>::value
+                         && is_comparable<typename T::second_type, U>::value
+                         && is_comparable<typename T::first_type, V>::value
+                         && is_comparable<typename T::second_type, V>::value
+                         && rocprim::is_arithmetic<typename T::first_type>::value
+                         && rocprim::is_arithmetic<typename T::second_type>::value
+                         && rocprim::is_arithmetic<U>::value && rocprim::is_arithmetic<V>::value),
+                        std::pair<T, T>>
+{
+
+    return {
+        T{limit_cast<typename T::first_type>(range_start),
+          limit_cast<typename T::second_type>(range_start)},
+        T{  limit_cast<typename T::first_type>(range_end),
+          limit_cast<typename T::second_type>(range_end)  }
+    };
+}
+
+template<class T, class U, class V>
+inline auto limit_random_range(U range_start, V range_end)
+    -> std::enable_if_t<!is_custom_type<T>::value && is_comparable<T, U>::value
+                            && is_comparable<T, V>::value,
+                        std::pair<T, T>>
+{
+
+    if(is_comparable<V, U>::value)
+    {
+        using common_type = typename std::common_type<T, U>::type;
+        if(static_cast<common_type>(range_start) > static_cast<common_type>(range_end))
+        {
+            throw std::range_error("limit_random_range: Incorrect range used!");
+        }
+    }
+
+    T start = limit_cast<T>(range_start);
+    T end   = limit_cast<T>(range_end);
+    return std::make_pair(start, end);
+}
+
 inline bool is_warp_size_supported(const unsigned int required_warp_size, const int device_id)
 {
     unsigned int warp_size;
@@ -306,21 +492,20 @@ inline bool is_warp_size_supported(const unsigned int required_warp_size, const
 }
 
 template<unsigned int LogicalWarpSize>
-struct DeviceSelectWarpSize
-{
-    static constexpr unsigned int value = ::rocprim::device_warp_size() >= LogicalWarpSize
-        ? LogicalWarpSize
-        : ::rocprim::device_warp_size();
-};
+__device__ constexpr bool device_test_enabled_for_warp_size_v
+    = ::rocprim::device_warp_size() >= LogicalWarpSize;
 
+/// \brief Get segments of uniform random size in [1, max_segment_length] with random key.
 template<typename T>
 std::vector<T>
-    get_random_segments(const size_t size, const size_t max_segment_length, const int seed_value)
+    get_random_segments(const size_t size, const size_t max_segment_length, unsigned int seed)
 {
     static_assert(rocprim::is_arithmetic<T>::value, "Key type must be arithmetic");
 
-    std::default_random_engine            prng(seed_value);
-    std::uniform_int_distribution<size_t> segment_length_distribution(max_segment_length);
+    engine_type                           prng(seed);
+    std::uniform_int_distribution<size_t> segment_length_distribution(
+        std::numeric_limits<size_t>::min(),
+        max_segment_length);
     // std::uniform_real_distribution cannot handle rocprim::half, use float instead
     using dis_type =
         typename std::conditional<std::is_same<rocprim::half, T>::value, float, T>::type;
@@ -342,6 +527,29 @@ std::vector<T>
     return keys;
 }
 
+/// \brief Get segments of uniform random size in [1, max_segment_length] with unique incrementing key.
+template<typename T>
+std::vector<T>
+    get_random_segments_iota(const size_t size, const size_t max_segment_length, unsigned int seed)
+{
+    engine_type                           prng(seed);
+    std::uniform_int_distribution<size_t> segment_length_distribution(1, max_segment_length);
+
+    std::vector<T> keys(size);
+
+    size_t segment_index    = 0;
+    size_t keys_start_index = 0;
+    while(keys_start_index < size)
+    {
+        const size_t new_segment_length = segment_length_distribution(prng);
+        const size_t new_segment_end    = std::min(size, keys_start_index + new_segment_length);
+        const T      key                = segment_index++;
+        std::fill(keys.begin() + keys_start_index, keys.begin() + new_segment_end, key);
+        keys_start_index += new_segment_length;
+    }
+    return keys;
+}
+
 template <typename T, T, typename>
 struct make_index_range_impl;
 
@@ -371,13 +579,16 @@ void static_for_each(Args&&... args)
                                                                  std::forward<Args>(args)...);
 }
 
-#define REGISTER_BENCHMARK(benchmarks, size, stream, instance)                     \
-    benchmark::internal::Benchmark* benchmark = benchmark::RegisterBenchmark(      \
-        instance.name().c_str(),                                                   \
-        [instance](benchmark::State& state, size_t size, const hipStream_t stream) \
-        { instance.run(state, size, stream); },                                    \
-        size,                                                                      \
-        stream);                                                                   \
+#define REGISTER_BENCHMARK(benchmarks, size, seed, stream, instance)                     \
+    benchmark::internal::Benchmark* benchmark = benchmark::RegisterBenchmark(            \
+        instance.name().c_str(),                                                         \
+        [instance](benchmark::State&   state,                                            \
+                   size_t              _size,                                            \
+                   const managed_seed& _seed,                                            \
+                   hipStream_t         _stream) { instance.run(state, _size, _seed, _stream); }, \
+        size,                                                                            \
+        seed,                                                                            \
+        stream);                                                                         \
     benchmarks.emplace_back(benchmark)
 
 struct config_autotune_interface
@@ -387,8 +598,8 @@ struct config_autotune_interface
     {
         return name();
     };
-    virtual ~config_autotune_interface()                           = default;
-    virtual void run(benchmark::State&, size_t, hipStream_t) const = 0;
+    virtual ~config_autotune_interface()                                                = default;
+    virtual void run(benchmark::State&, size_t, const managed_seed&, hipStream_t) const = 0;
 };
 
 struct config_autotune_register
@@ -413,10 +624,11 @@ struct config_autotune_register
 
     // Register a subset of all created benchmarks for the current parallel instance and add to vector.
     static void register_benchmark_subset(std::vector<benchmark::internal::Benchmark*>& benchmarks,
-                                          int               parallel_instance_index,
-                                          int               parallel_instance_count,
-                                          size_t            size,
-                                          const hipStream_t stream)
+                                          int                 parallel_instance_index,
+                                          int                 parallel_instance_count,
+                                          size_t              size,
+                                          const managed_seed& seed,
+                                          hipStream_t         stream)
     {
         std::vector<std::unique_ptr<config_autotune_interface>>& configs = vector();
         // sorting to get a consistent order because order of initialization of static variables is undefined by the C++ standard.
@@ -433,9 +645,13 @@ struct config_autotune_register
             config_autotune_interface*                  tuning_benchmark = uniq_ptr.get();
             benchmark::internal::Benchmark*             benchmark = benchmark::RegisterBenchmark(
                 tuning_benchmark->name().c_str(),
-                [tuning_benchmark](benchmark::State& state, size_t size, const hipStream_t stream)
-                { tuning_benchmark->run(state, size, stream); },
+                [tuning_benchmark](benchmark::State&   state,
+                                   size_t              size,
+                                   const managed_seed& seed,
+                                   hipStream_t         stream)
+                { tuning_benchmark->run(state, size, seed, stream); },
                 size,
+                seed,
                 stream);
             benchmarks.emplace_back(benchmark);
         }
@@ -642,6 +858,11 @@ struct Traits
 };
 
 // Explicit definitions
+template<>
+inline const char* Traits<char>::name()
+{
+    return "char";
+}
 template <>
 inline const char* Traits<int>::name() { return "int"; }
 template <>
@@ -695,6 +916,11 @@ inline const char* Traits<custom_type<double, double>>::name()
     return "custom_type<double,double>";
 }
 template<>
+inline const char* Traits<custom_type<int, double>>::name()
+{
+    return "custom_type<int,double>";
+}
+template<>
 inline const char* Traits<custom_type<char, double>>::name()
 {
     return "custom_type<char,double>";
@@ -715,6 +941,11 @@ inline const char* Traits<custom_type<long long, double>>::name()
     return "custom_type<int64_t,double>";
 }
 template<>
+inline const char* Traits<custom_type<float, int16_t>>::name()
+{
+    return "custom_type<float,int16_t>";
+}
+template<>
 inline const char* Traits<rocprim::empty_type>::name()
 {
     return "empty_type";
@@ -846,4 +1077,18 @@ struct alignas(Alignment) custom_aligned_type
     unsigned char data[Size];
 };
 
+template<typename Config>
+std::string partition_config_name()
+{
+    const rocprim::detail::partition_config_params config = Config();
+    return "{bs:" + std::to_string(config.kernel_config.block_size)
+           + ",ipt:" + std::to_string(config.kernel_config.items_per_thread) + "}";
+}
+
+template<>
+inline std::string partition_config_name<rocprim::default_config>()
+{
+    return "default_config";
+}
+
 #endif // ROCPRIM_BENCHMARK_UTILS_HPP_
