@@ -1,6 +1,6 @@
 // MIT License
 //
-// Copyright (c) 2017-2023 Advanced Micro Devices, Inc. All rights reserved.
+// Copyright (c) 2017-2024 Advanced Micro Devices, Inc. All rights reserved.
 //
 // Permission is hereby granted, free of charge, to any person obtaining a copy
 // of this software and associated documentation files (the "Software"), to deal
@@ -30,21 +30,22 @@
 #include <rocprim/iterator/transform_iterator.hpp>
 
 // required test headers
+#include "rocprim/types.hpp"
 #include "test_utils_custom_test_types.hpp"
 #include "test_utils_types.hpp"
 
-template<
-    class Key,
-    class Value,
-    class ReduceOp,
-    unsigned int MinSegmentLength,
-    unsigned int MaxSegmentLength,
-    class Aggregate = Value,
-    class KeyCompareFunction = ::rocprim::equal_to<Key>,
-    // Tests output iterator with void value_type (OutputIterator concept)
-    bool UseIdentityIterator = false,
-    bool UseGraphs = false
->
+template<class Key,
+         class Value,
+         class ReduceOp,
+         unsigned int MinSegmentLength,
+         unsigned int MaxSegmentLength,
+         class Aggregate          = Value,
+         class KeyCompareFunction = ::rocprim::equal_to<Key>,
+         // Tests output iterator with void value_type (OutputIterator concept)
+         bool UseIdentityIterator = false,
+         bool UseGraphs           = false,
+         bool Deterministic       = false,
+         typename Config          = rocprim::default_config>
 struct params
 {
     using key_type = Key;
@@ -56,6 +57,8 @@ struct params
     using key_compare_op = KeyCompareFunction;
     static constexpr bool use_identity_iterator = UseIdentityIterator;
     static constexpr bool use_graphs = UseGraphs;
+    static constexpr bool deterministic              = Deterministic;
+    using config                                     = Config;
 };
 
 template<class Params>
@@ -90,16 +93,21 @@ using custom_double2 = test_utils::custom_test_type<double>;
 // clang-format off
 typedef ::testing::Types<
     params<int, int, rocprim::plus<int>, 1, 1, int, rocprim::equal_to<int>, true>,
-    params<double, int, rocprim::plus<int>, 3, 5, long long, custom_key_compare_op1<double>>,
+    params<double, int, rocprim::plus<int>, 3, 5, long long, custom_key_compare_op1<double>, false, false, true>,
     params<float, custom_double2, rocprim::minimum<custom_double2>, 1, 10000>,
     params<custom_double2, custom_int2, rocprim::plus<custom_int2>, 1, 10>,
     params<unsigned long long, float, rocprim::minimum<float>, 1, 30>,
-    params<int, rocprim::half, rocprim::minimum<rocprim::half>, 15, 100>,
+    // with block size different than ROCPRIM_DEFAULT_MAX_BLOCK_SIZE and non-power-of-two items per thread
+    params<int, rocprim::half, rocprim::minimum<rocprim::half>, 15, 100, rocprim::half, rocprim::equal_to<int>, false, false, true, rocprim::reduce_by_key_config<128, 3>>,
+    // half should be supported, but is missing some key operators.
+    // we should uncomment these, as soon as these are implemented and the tests compile and work as intended.
+    //params<rocprim::half, rocprim::half, rocprim::minimum<rocprim::half>, 15, 100>,
     params<int, rocprim::bfloat16, rocprim::minimum<rocprim::bfloat16>, 15, 100>,
+    params<rocprim::bfloat16, rocprim::bfloat16, rocprim::minimum<rocprim::bfloat16>, 15, 100>,
     params<int, unsigned int, rocprim::maximum<unsigned int>, 20, 100>,
     params<float, long long, rocprim::maximum<unsigned long long>, 100, 400, long long, custom_key_compare_op1<float>>,
     params<unsigned int, unsigned char, rocprim::plus<unsigned char>, 200, 600>,
-    params<double, int, rocprim::plus<int>, 100, 2000, double, custom_key_compare_op1<double>>,
+    params<double, int, rocprim::plus<int>, 100, 2000, double, custom_key_compare_op1<double>, false, false, true>,
     params<int8_t, int8_t, rocprim::maximum<int8_t>, 20, 100>,
     params<uint8_t, uint8_t, rocprim::maximum<uint8_t>, 20, 100>,
     params<char, rocprim::half, rocprim::maximum<rocprim::half>, 123, 1234>,
@@ -110,10 +118,23 @@ typedef ::testing::Types<
     params<unsigned int, double, rocprim::minimum<double>, 1000, 50000>,
     params<unsigned long long, unsigned long long, rocprim::plus<unsigned long long>, 100000, 100000>,
     params<test_utils::custom_test_array_type<double, 8>, unsigned long, rocprim::plus<>, 69, 420>,
-    params<int, int, rocprim::plus<int>, 1, 10, int, ::rocprim::equal_to<int>, false, true>    
+    params<int, int, rocprim::plus<int>, 1, 10, int, ::rocprim::equal_to<int>, false, true>
 > Params;
 // clang-format on
 
+template<bool Deterministic, typename Config = rocprim::default_config, typename... Args>
+constexpr hipError_t invoke_reduce_by_key(Args&&... args)
+{
+    if(Deterministic)
+    {
+        return rocprim::deterministic_reduce_by_key<Config>(std::forward<Args>(args)...);
+    }
+    else
+    {
+        return rocprim::reduce_by_key<Config>(std::forward<Args>(args)...);
+    }
+}
+
 TYPED_TEST_SUITE(RocprimDeviceReduceByKey, Params);
 
 TYPED_TEST(RocprimDeviceReduceByKey, ReduceByKey)
@@ -134,14 +155,14 @@ TYPED_TEST(RocprimDeviceReduceByKey, ReduceByKey)
         typename std::conditional<
             test_utils::is_valid_for_int_distribution<key_inner_type>::value,
             std::uniform_int_distribution<key_inner_type>,
-            typename std::conditional<std::is_signed<key_inner_type>::value,
-                std::uniform_int_distribution<int>,
-                std::uniform_int_distribution<unsigned int>
-            >::type
-        >::type
-    >::type;
+            typename std::conditional<rocprim::is_signed<key_inner_type>::value,
+                                      std::uniform_int_distribution<int>,
+                                      std::uniform_int_distribution<unsigned int>>::type>::type>::
+        type;
+    using config = typename TestFixture::params::config;
 
     constexpr bool use_identity_iterator = TestFixture::params::use_identity_iterator;
+    constexpr bool deterministic         = TestFixture::params::deterministic;
     const bool debug_synchronous = false;
 
     reduce_op_type reduce_op;
@@ -183,8 +204,8 @@ TYPED_TEST(RocprimDeviceReduceByKey, ReduceByKey)
             std::vector<value_type> values_input = test_utils::get_random_data<value_type>(size, 0, 100, seed_value);
 
             size_t offset = 0;
-            key_type prev_key = key_distribution_type(0, 100)(gen);
-            key_type current_key = prev_key + key_delta_dis(gen);
+            key_type prev_key    = static_cast<key_type>(key_distribution_type(0, 100)(gen));
+            key_type current_key = static_cast<key_type>(prev_key + key_delta_dis(gen));
             while(offset < size)
             {
                 const size_t key_count = key_count_dis(gen);
@@ -256,48 +277,50 @@ TYPED_TEST(RocprimDeviceReduceByKey, ReduceByKey)
 
             size_t temporary_storage_bytes;
 
-            hipGraph_t graph;
-            hipGraphExec_t graph_instance;
-            if (TestFixture::params::use_graphs)
-                graph = test_utils::createGraphHelper(stream);
-            
-            HIP_CHECK(
-                rocprim::reduce_by_key(
-                    nullptr, temporary_storage_bytes,
-                    d_keys_input, d_values_input, size,
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_unique_output),
-                    test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_aggregates_output),
-                    d_unique_count_output,
-                    reduce_op, key_compare_op,
-                    stream, debug_synchronous
-                )
-            );
+            HIP_CHECK((invoke_reduce_by_key<deterministic, config>(
+                nullptr,
+                temporary_storage_bytes,
+                d_keys_input,
+                d_values_input,
+                size,
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_unique_output),
+                test_utils::wrap_in_identity_iterator<use_identity_iterator>(d_aggregates_output),
+                d_unique_count_output,
+                reduce_op,
+                key_compare_op,
+                stream,
+                debug_synchronous)));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
             ASSERT_GT(temporary_storage_bytes, 0);
 
             void * d_temporary_storage;
             HIP_CHECK(test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
 
-            if (TestFixture::params::use_graphs)
-                test_utils::resetGraphHelper(graph, graph_instance, stream);
-            
-            HIP_CHECK(
-                rocprim::reduce_by_key(
-                    d_temporary_storage, temporary_storage_bytes,
-                    d_keys_input, d_values_input, size,
-                    d_unique_output, d_aggregates_output,
-                    d_unique_count_output,
-                    reduce_op, key_compare_op,
-                    stream, debug_synchronous
-                )
-            );
+            test_utils::GraphHelper gHelper;
+            if(TestFixture::params::use_graphs)
+            {
+               gHelper.startStreamCapture(stream);
+            }
+
+            HIP_CHECK((invoke_reduce_by_key<deterministic, config>(d_temporary_storage,
+                                                                   temporary_storage_bytes,
+                                                                   d_keys_input,
+                                                                   d_values_input,
+                                                                   size,
+                                                                   d_unique_output,
+                                                                   d_aggregates_output,
+                                                                   d_unique_count_output,
+                                                                   reduce_op,
+                                                                   key_compare_op,
+                                                                   stream,
+                                                                   debug_synchronous)));
 
-            if (TestFixture::params::use_graphs)
-                graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
             
+            if(TestFixture::params::use_graphs)
+            {
+                gHelper.createAndLaunchGraph(stream);
+            }
+
             HIP_CHECK(hipFree(d_temporary_storage));
 
             std::vector<key_type> unique_output(unique_count_expected);
@@ -333,10 +356,10 @@ TYPED_TEST(RocprimDeviceReduceByKey, ReduceByKey)
 
             if (TestFixture::params::use_graphs)
             {
-                test_utils::cleanupGraphHelper(graph, graph_instance);
+                gHelper.cleanupGraphHelper();
                 HIP_CHECK(hipStreamDestroy(stream));
             }
-            
+
             ASSERT_EQ(unique_count_output[0], unique_count_expected);
 
             ASSERT_NO_FATAL_FAILURE(test_utils::assert_eq(unique_output, unique_expected));
@@ -345,7 +368,7 @@ TYPED_TEST(RocprimDeviceReduceByKey, ReduceByKey)
     }
 }
 
-template<typename value_type, bool use_graphs = false>
+template<typename value_type, bool use_graphs = false, bool Deterministic = false>
 void large_indices_reduce_by_key()
 {
     int device_id = test_common_utils::obtain_device_from_ctest();
@@ -401,51 +424,49 @@ void large_indices_reduce_by_key()
 
         size_t temporary_storage_bytes;
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-        
-        HIP_CHECK(rocprim::reduce_by_key(nullptr,
-                                         temporary_storage_bytes,
-                                         d_keys_input,
-                                         d_values_input,
-                                         size,
-                                         d_unique_output,
-                                         d_aggregates_output,
-                                         d_unique_count_output,
-                                         reduce_op,
-                                         key_compare_op,
-                                         stream,
-                                         debug_synchronous));
+        HIP_CHECK(invoke_reduce_by_key<Deterministic>(nullptr,
+                                                      temporary_storage_bytes,
+                                                      d_keys_input,
+                                                      d_values_input,
+                                                      size,
+                                                      d_unique_output,
+                                                      d_aggregates_output,
+                                                      d_unique_count_output,
+                                                      reduce_op,
+                                                      key_compare_op,
+                                                      stream,
+                                                      debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-        
         ASSERT_GT(temporary_storage_bytes, 0);
 
         void* d_temporary_storage;
         HIP_CHECK(
             test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
+        test_utils::GraphHelper gHelper;
+        if(use_graphs)
+        {
+           gHelper.startStreamCapture(stream);
+        }
 
-        HIP_CHECK(rocprim::reduce_by_key(d_temporary_storage,
-                                         temporary_storage_bytes,
-                                         d_keys_input,
-                                         d_values_input,
-                                         size,
-                                         d_unique_output,
-                                         d_aggregates_output,
-                                         d_unique_count_output,
-                                         reduce_op,
-                                         key_compare_op,
-                                         stream,
-                                         debug_synchronous));
+        HIP_CHECK(invoke_reduce_by_key<Deterministic>(d_temporary_storage,
+                                                      temporary_storage_bytes,
+                                                      d_keys_input,
+                                                      d_values_input,
+                                                      size,
+                                                      d_unique_output,
+                                                      d_aggregates_output,
+                                                      d_unique_count_output,
+                                                      reduce_op,
+                                                      key_compare_op,
+                                                      stream,
+                                                      debug_synchronous));
 
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+        
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
 
         HIP_CHECK(hipFree(d_temporary_storage));
 
@@ -469,9 +490,11 @@ void large_indices_reduce_by_key()
         HIP_CHECK(hipFree(d_aggregates_output));
         HIP_CHECK(hipFree(d_unique_count_output));
 
-        if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
-        
+        if(use_graphs)
+        {
+            gHelper.cleanupGraphHelper();
+        }
+
         ASSERT_EQ(unique_count_output[0], unique_count_expected);
 
         size_t total_size = 0;
@@ -488,8 +511,10 @@ void large_indices_reduce_by_key()
         ASSERT_EQ(value_type(size - total_size), aggregates_output[last_idx]);
     }
 
-    if (use_graphs)
+    if(use_graphs)
+    {
         HIP_CHECK(hipStreamDestroy(stream));
+    }
 }
 
 TEST(RocprimDeviceReduceByKey, LargeIndicesReduceByKeySmallValueType)
@@ -509,7 +534,12 @@ TEST(RocprimDeviceReduceByKey, LargeIndicesReduceByKeyLargeValueTypeWithGraphs)
     large_indices_reduce_by_key<test_utils::custom_test_type<size_t>, true>();
 }
 
-template<typename value_type, bool use_graphs = false>
+TEST(RocprimDeviceReduceByKey, LargeIndicesReduceByKeyDeterministic)
+{
+    large_indices_reduce_by_key<double, false, true>();
+}
+
+template<typename value_type, bool use_graphs = false, bool Deterministic = false>
 void large_segment_count_reduce_by_key()
 {
     int device_id = test_common_utils::obtain_device_from_ctest();
@@ -548,51 +578,51 @@ void large_segment_count_reduce_by_key()
         HIP_CHECK(test_common_utils::hipMallocHelper(&d_unique_count_output,
                                                      sizeof(*d_unique_count_output)));
 
-        hipGraph_t graph;
-        hipGraphExec_t graph_instance;
-        if (use_graphs)
-            graph = test_utils::createGraphHelper(stream);
-            
         size_t temporary_storage_bytes;
-        HIP_CHECK(rocprim::reduce_by_key(nullptr,
-                                         temporary_storage_bytes,
-                                         d_keys_input,
-                                         d_values_input,
-                                         size,
-                                         d_unique_output,
-                                         d_aggregates_output,
-                                         d_unique_count_output,
-                                         reduce_op,
-                                         key_compare_op,
-                                         stream,
-                                         debug_synchronous));
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
-            
+        HIP_CHECK(invoke_reduce_by_key<Deterministic>(nullptr,
+                                                      temporary_storage_bytes,
+                                                      d_keys_input,
+                                                      d_values_input,
+                                                      size,
+                                                      d_unique_output,
+                                                      d_aggregates_output,
+                                                      d_unique_count_output,
+                                                      reduce_op,
+                                                      key_compare_op,
+                                                      stream,
+                                                      debug_synchronous));
+
         ASSERT_GT(temporary_storage_bytes, 0);
 
         void* d_temporary_storage;
         HIP_CHECK(
             test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
 
-        if (use_graphs)
-            test_utils::resetGraphHelper(graph, graph_instance, stream);
-                
-        HIP_CHECK(rocprim::reduce_by_key(d_temporary_storage,
-                                         temporary_storage_bytes,
-                                         d_keys_input,
-                                         d_values_input,
-                                         size,
-                                         d_unique_output,
-                                         d_aggregates_output,
-                                         d_unique_count_output,
-                                         reduce_op,
-                                         key_compare_op,
-                                         stream,
-                                         debug_synchronous));
-        if (use_graphs)
-            graph_instance = test_utils::endCaptureGraphHelper(graph, stream, true, true);
+        test_utils::GraphHelper gHelper;
+        if(use_graphs)
+        {
+           gHelper.startStreamCapture(stream);
+        }
+
+        HIP_CHECK(invoke_reduce_by_key<Deterministic>(d_temporary_storage,
+                                                      temporary_storage_bytes,
+                                                      d_keys_input,
+                                                      d_values_input,
+                                                      size,
+                                                      d_unique_output,
+                                                      d_aggregates_output,
+                                                      d_unique_count_output,
+                                                      reduce_op,
+                                                      key_compare_op,
+                                                      stream,
+                                                      debug_synchronous));
+
         
+        if(use_graphs)
+        {
+            gHelper.createAndLaunchGraph(stream);
+        }
+
         HIP_CHECK(hipFree(d_temporary_storage));
 
         size_t unique_count_output;
@@ -606,7 +636,7 @@ void large_segment_count_reduce_by_key()
         ASSERT_EQ(unique_count_output, unique_count_expected);
 
         if (use_graphs)
-            test_utils::cleanupGraphHelper(graph, graph_instance);
+            gHelper.cleanupGraphHelper();
     }
 
     if (use_graphs)
@@ -629,6 +659,11 @@ TEST(RocprimDeviceReduceByKey, GraphReduceByKey)
     large_segment_count_reduce_by_key<unsigned int, true>();
 }
 
+TEST(RocprimDeviceReduceByKey, LargeSegmentCountReduceByKeyDeterministic)
+{
+    large_segment_count_reduce_by_key<float, false, true>();
+}
+
 TEST(RocprimDeviceReduceByKey, ReduceByNonEqualKeys)
 {
     int device_id = test_common_utils::obtain_device_from_ctest();
@@ -639,6 +674,7 @@ TEST(RocprimDeviceReduceByKey, ReduceByNonEqualKeys)
     using value_type = unsigned int;
 
     const bool debug_synchronous = false;
+    constexpr bool deterministic     = false;
 
     ::rocprim::plus<value_type> reduce_op;
     auto                        key_compare_op = [](const auto&, const auto&) { return false; };
@@ -674,18 +710,18 @@ TEST(RocprimDeviceReduceByKey, ReduceByNonEqualKeys)
 
             size_t temporary_storage_bytes;
 
-            HIP_CHECK(rocprim::reduce_by_key(nullptr,
-                                             temporary_storage_bytes,
-                                             d_keys_input,
-                                             d_values_input,
-                                             size,
-                                             d_unique_output,
-                                             d_aggregates_output,
-                                             d_unique_count_output,
-                                             reduce_op,
-                                             key_compare_op,
-                                             stream,
-                                             debug_synchronous));
+            HIP_CHECK(invoke_reduce_by_key<deterministic>(nullptr,
+                                                          temporary_storage_bytes,
+                                                          d_keys_input,
+                                                          d_values_input,
+                                                          size,
+                                                          d_unique_output,
+                                                          d_aggregates_output,
+                                                          d_unique_count_output,
+                                                          reduce_op,
+                                                          key_compare_op,
+                                                          stream,
+                                                          debug_synchronous));
 
             ASSERT_GT(temporary_storage_bytes, 0);
 
@@ -693,18 +729,18 @@ TEST(RocprimDeviceReduceByKey, ReduceByNonEqualKeys)
             HIP_CHECK(
                 test_common_utils::hipMallocHelper(&d_temporary_storage, temporary_storage_bytes));
 
-            HIP_CHECK(rocprim::reduce_by_key(d_temporary_storage,
-                                             temporary_storage_bytes,
-                                             d_keys_input,
-                                             d_values_input,
-                                             size,
-                                             d_unique_output,
-                                             d_aggregates_output,
-                                             d_unique_count_output,
-                                             reduce_op,
-                                             key_compare_op,
-                                             stream,
-                                             debug_synchronous));
+            HIP_CHECK(invoke_reduce_by_key<deterministic>(d_temporary_storage,
+                                                          temporary_storage_bytes,
+                                                          d_keys_input,
+                                                          d_values_input,
+                                                          size,
+                                                          d_unique_output,
+                                                          d_aggregates_output,
+                                                          d_unique_count_output,
+                                                          reduce_op,
+                                                          key_compare_op,
+                                                          stream,
+                                                          debug_synchronous));
 
             HIP_CHECK(hipFree(d_temporary_storage));
 
